/*
 * Copyright 2011 Steven Gribble
 *
 *  This file is part of the UW CSE 333 course project sequence
 *  (333proj).
 *
 *  333proj is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  333proj is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with 333proj.  If not, see <http://www.gnu.org/licenses/>.
 */

#include <stdio.h>
#include <stdlib.h>
#include <assert.h>

#include "ll.h"
#include "ll_priv.h"

LinkedList AllocateLinkedList(void) {
  LinkedListRecordPtr ll;

  ll = (LinkedListRecordPtr) malloc(sizeof(LinkedListRecord));
  if (ll == NULL) {
    // out of memory
    return NULL;
  }

  // initialize the newly allocated record structure
  // STEP 1.



  // return our newly minted linked list
  return (void *) ll;
}

void FreeLinkedList(LinkedList list,
                    PayloadFreeFnPtr payload_free_function) {
  LinkedListRecordPtr ll = (LinkedListRecordPtr) list;

  assert(ll != NULL);  // defensive programming is a good idea
  assert(payload_free_function != NULL);

  // sweep through the list and free all of the nodes' payloads as
  // well as the nodes themselves
  while (ll->head != NULL) {
    // STEP 2.


  }

  // free the list record itself
  free(ll);
}

unsigned int NumElementsInLinkedList(LinkedList list) {
  LinkedListRecordPtr ll = (LinkedListRecordPtr) list;

  assert(ll != NULL);  // defensive programming
  return ll->num_elements;
}

int PushLinkedList(LinkedList list, void *payload) {
  LinkedListRecordPtr ll = (LinkedListRecordPtr) list;
  LinkedListNodePtr ln = NULL;

  assert(ll != NULL);  // defensive programming

  // allocate space for the new node.
  // STEP 3.



  // set the payload
  ln->payload = payload;

  if (ll->num_elements == 0) {
    // degenerate case; list is currently empty
    assert(ll->head == NULL);  // debugging aid
    assert(ll->tail == NULL);  // debugging aid
    ln->next = ln->prev = NULL;
    ll->head = ll->tail = ln;
    ll->num_elements = 1;
  } else {
    // typical case; list has >=1 elements
    // STEP 4.


  }

  // return success
  return 1;
}

int PopLinkedList(LinkedList list, void **payload_ptr) {
  LinkedListRecordPtr ll = (LinkedListRecordPtr) list;

  // defensive programming
  assert(payload_ptr != NULL);
  assert(ll != NULL);

  // if there is nothing in the list, fail.
  if (ll->num_elements == 0) {
    return 0;
  }

  // There are two cases you probably want to
  // consider: (a) a list with a single element in it,
  // (b) a list with >=2 elements in it.
  //
  // STEP 5.



  // depending on what you did in step 5, you might need
  // to change this return to do something else.
  return 0;
}

int AppendLinkedList(LinkedList list, void *payload) {
  LinkedListRecordPtr ll = (LinkedListRecordPtr) list;
  LinkedListNodePtr ln = NULL;

  assert(ll != NULL);  // defensive programming

  // attempt to allocate space for the new node.
  // STEP 6.


  // set the payload
  ln->payload = payload;

  if (ll->num_elements == 0) {
    // degenerate case; list is currently empty
    assert(ll->head == NULL);  // debugging aid
    assert(ll->tail == NULL);  // debugging aid
    ln->next = ln->prev = NULL;
    ll->head = ll->tail = ln;
    ll->num_elements = 1;
  } else {
    // typical case; list has >=1 elements
    // STEP 7.


  }

  // return success
  return 1;
}

int SliceLinkedList(LinkedList list, void **payload_ptr) {
  LinkedListRecordPtr ll = (LinkedListRecordPtr) list;

  // defensive programming
  assert(payload_ptr != NULL);
  assert(ll != NULL);

  // if there is nothing in the list, fail.
  if (ll->num_elements == 0) {
    return 0;
  }

  // There are two cases you probably want to
  // consider: (a) a list with a single element in it,
  // (b) a list with >=2 elements in it.
  //
  // STEP 8.


  // depending on what you did in step 5, you might need
  // to change this return to do something else.
  return 0;
}

void SortLinkedList(LinkedList list, unsigned int ascending,
                    PayloadComparatorFnPtr comparator_function) {
  LinkedListRecordPtr ll = (LinkedListRecordPtr) list;
  int                 swapped;

  assert(ll != NULL);  // defensive programming

  if (ll->num_elements < 2) {
    // no sorting needed
    return;
  }

  // we'll implement bubblesort!  nice and easy, and nice and slow :)
  do {
    LinkedListNodePtr curnode;

    swapped = 0;
    curnode = ll->head;
    while (curnode->next != NULL) {
      int compare_result = comparator_function(curnode->payload,
                                               curnode->next->payload);
      if (ascending) {
        compare_result *= -1;
      }
      if (compare_result < 0) {
        // bubble-swap payloads
        void *tmp;
        tmp = curnode->payload;
        curnode->payload = curnode->next->payload;
        curnode->next->payload = tmp;
        swapped = 1;
      }
      curnode = curnode->next;
    }
  } while (swapped);
}

LLIter LinkedListMakeIterator(LinkedList list, int pos) {
  LinkedListRecordPtr ll = (LinkedListRecordPtr) list;
  LLIterRecordPtr li;

  // defensive programming
  assert(ll != NULL);
  assert((pos == 0) || (pos == 1));

  li = (LLIterRecordPtr) malloc(sizeof(LLIterRecord));
  if (li == NULL) {
    // out of memory!
    return NULL;
  }

  // set up the iterator
  li->list = ll;
  if (pos == 0) {
    li->node = ll->head;
  } else {
    li->node = ll->tail;
  }

  // return the new iterator
  return (void *) li;
}

void LLIteratorFree(LLIter iter) {
  LLIterRecordPtr li = (LLIterRecordPtr) iter;

  assert(li != NULL);  // defensive programming
  free(li);
}

int LLIteratorNext(LLIter iter) {
  LLIterRecordPtr li = (LLIterRecordPtr) iter;

  assert(li != NULL);  // defensive programming

  // advance the iterator; there are multiple cases
  // to consider.
  // STEP 9.



  return 1;
}

int LLIteratorAtTail(LLIter iter) {
  LLIterRecordPtr li = (LLIterRecordPtr) iter;

  assert(li != NULL);  // defensive programming

  // test to see if we're at the end of the list.
  // are there any degenerate cases to consider?
  // STEP 10.



  return 0;
}

int LLIteratorPrev(LLIter iter) {
  LLIterRecordPtr li = (LLIterRecordPtr) iter;

  assert(li != NULL);  // defensive programming

  // rewind the iterator; there are multiple cases
  // to consider.
  // STEP 11.


  return 1;
}

int LLIteratorAtHead(LLIter iter) {
  LLIterRecordPtr li = (LLIterRecordPtr) iter;

  assert(li != NULL);  // defensive programming

  // test to see if we're at the end of the list.
  // are there any degenerate cases to consider?
  // STEP 12.


  return 0;
}

int LLIteratorGetPayload(LLIter iter, void **payload) {
  LLIterRecordPtr li = (LLIterRecordPtr) iter;

  assert(li != NULL);  // defensive programming

  // return (through the return parameter "payload")
  // the payload that iter points to.
  // STEP 13.


  return 1;
}

int LLIteratorDelete(LLIter iter,
                     PayloadFreeFnPtr payload_free_function) {
  LLIterRecordPtr li = (LLIterRecordPtr) iter;

  // defensive programming
  assert(li != NULL);
  assert(payload_free_function != NULL);
  if (li->node == NULL) {
    return -1;
  }

  // delete the list node the iterator points to, being
  // sure to free the payload.

  // degenerate case: list has a single element.
  if (li->list->num_elements == 1) {
    // use our own PopLinkedList function
    void *payload;
    PopLinkedList((LinkedList) li->list, &payload);
    payload_free_function(payload);
    li->node = NULL;
    return 0;
  }

  // figure out the other cases and implement them.
  // be careful to not leak memory.
  // STEP 14.



  return -1;
}

int LLIteratorInsertBefore(LLIter iter,
                           void *payload) {
  LLIterRecordPtr li = (LLIterRecordPtr) iter;

  assert(li != NULL);  // defensive programming!

  // degenerate case: list is empty, or cursor
  // is pointing at the head, or cursor not pointing
  // at anything.  Use our PushLinkedList function.
  if ((li->node == NULL) ||
      (li->list->num_elements <= 1) ||
      (li->node == li->list->head)) {
    int retval = PushLinkedList((LinkedList) li->list,
                                payload);
    if (retval == 1) {
      // if we just added our first node, need
      // to set the iterator to point to it, otherwise
      // don't modify the iterator.
      if (li->list->num_elements == 1) {
        li->node = li->list->head;
      }
    }
    return retval;
  }

  // figure out the general case and implement it.
  // be sure to read the ll.h header file to make sure
  // you understand the semantics of the operation.
  // STEP 15.



  return 0;
}
